from .value import Value, DecimalValue, RationalValue
from .definition import (
    Definition,
    NewValues,
    InSingleDefinition,
    InPairDefinition,
    DomainDefinition,
)
from .error import ParseError
from .compose import DefinitionSet
from .calculator import Calculator



from abc import ABC, abstractmethod


class Value(ABC):
    @abstractmethod
    def get_common_representation(self) -> str:
        pass

    @abstractmethod
    def get_formula_representation(self) -> str:
        pass

    @abstractmethod
    def cmp_to_user_input(self, s: str) -> bool:
        pass

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {self.get_common_representation()}>"



from .abc import Definition, NewValues
from damn.value import Value, DecimalValue
import re
import typing
import random


class InSingleDefinition(Definition):
    pattern = re.compile(r"(\w+) in \[(.+)]")

    def __init__(self, name: str, values: typing.List[Value]):
        self.name = name
        self.values = values

    @classmethod
    def parse(cls, s: str) -> typing.Optional["InSingleDefinition"]:
        m = re.match(cls.pattern, s)
        if not m:
            return None
        m = m.groups()
        values = []
        for value in m[1].split(","):
            values.append(DecimalValue.from_str(value.rstrip(" ")))
        return cls(m[0], values)

    def generate(self) -> NewValues:
        return {self.name: random.choice(self.values)}



from .abc import Definition, NewValues
from damn.value import DecimalValue
import re
import typing
import random


class DomainDefinition(Definition):
    pattern = re.compile(r"(.+) > (\w+) > (.+)")

    def __init__(
        self,
        name: str,
        start: DecimalValue,
        end: DecimalValue,
        is_integer: bool = False,
    ):
        self.name = name
        self.start = start
        self.end = end
        self.is_integer = is_integer

    @classmethod
    def parse(cls, s: str) -> typing.Optional["DomainDefinition"]:
        m = re.match(cls.pattern, s)
        if not m:
            return None
        m = m.groups()
        start, name, end = list(m)
        start = DecimalValue.from_str(start)
        end = DecimalValue.from_str(end)
        return cls(name, start, end, is_integer=False)

    def generate(self) -> NewValues:
        if self.is_integer:
            result = random.randint(self.start.v, self.end.v)
        else:
            result = random.uniform(self.start.v, self.end.v)
        result = round(result, 2)
        return {self.name: DecimalValue(result)}



from .abc import Value
import typing


class DecimalValue(Value):
    def __init__(self, v: typing.Union[int, float]):
        self.v = v

    def get_common_representation(self) -> str:
        return str(self.v)

    def get_formula_representation(self) -> str:
        return str(self.v)

    @classmethod
    def from_str(cls, s: str) -> "DecimalValue":
        return cls(int(s) if s.isdigit() else float(s))

    def cmp_to_user_input(self, s: str) -> bool:
        s = s.replace(",", ".")
        return (
            s == self.get_common_representation()
            or s == self.get_formula_representation()
        )



import itertools
from .value import Value, DecimalValue
from .lib import wolfram_calculate


def convert_kv(s: str) -> tuple[str, float]:
    s = s.strip().split(" equals ")
    n, v = s
    v = float(v.replace("minus ", "-").replace("negative ", "-"))
    return n, round(v, 2)


class Calculator:
    def __init__(self, apps: list[str]):
        self.apps = itertools.cycle(apps)

    async def calculate(self, data: dict[str, Value], formula: str) -> DecimalValue:
        s = ", ".join(
            (f"{k} = {v.get_formula_representation()}" for k, v in data.items())
        )
        result = await wolfram_calculate(next(self.apps), s + ", " + formula)
        variables = dict(
            (
                convert_kv(v)
                for v in result.replace(
                    " and ",
                    "",
                ).split(",")
            )
        )
        for variable in data:
            variables.pop(variable)
        return DecimalValue(list(variables.values())[0])



import typing

if typing.TYPE_CHECKING:
    from .definition.abc import Definition


class ParseError(Exception):
    def __init__(self, t: typing.Type["Definition"], err: str = ""):
        self.t = t
        self.err = err

    def __repr__(self) -> str:
        return "(ParseError in {}: {})".format(self.t.__name__, self.err)



import aiohttp

API_URL = "http://api.wolframalpha.com/v1"


async def wolfram_calculate(app: str, query: str) -> str:
    async with aiohttp.ClientSession() as session:
        result = await session.get(
            API_URL + "/result", params={"appid": app, "i": query}
        )
        return await result.text()
